# -*- coding: utf-8 -*-

from nawano.exceptions import ValidationError, NoRecordsFound
from nawano.utils import render_table, stylize
from .state import StateService


class NawanoService(object):
    __model__ = None
    __state__ = StateService()

    def get_one(self, raise_on_empty=False, **kwargs):
        obj = self.__model__.query(**kwargs).one_or_none()

        if not obj and raise_on_empty:
            raise NoRecordsFound

        return obj

    def get_many(self, raise_on_empty=False, **kwargs):
        objs = self.__model__.query(**kwargs).all()

        if not objs and raise_on_empty:
            raise NoRecordsFound

        return objs

    def insert(self, **data):
        return self.__model__.insert(**data)

    def update(self, _id, **data):
        return self.__model__.update(_id, **data)

    def delete(self, **kwargs):
        return self.__model__.delete(**kwargs)

    @staticmethod
    def validate_name(name, argument_name='name'):
        if not 2 <= len(name) <= 30:
            raise ValidationError('argument {0} must contain between {1}'
                                  .format(argument_name, stylize('2 and 30 characters', bold=True)))
        elif ' ' in name:
            raise ValidationError('may {0}'.format(stylize('not contain spaces', bold=True)))

        return name

    def validate_address(self, address):
        return self.__state__.network.validate_address(address)

    @staticmethod
    def funds_text(funds):
        return '\n{0}available: {1} \n{0}pending: {2}'.format(' '*2, funds['available'], funds['pending'])

    @staticmethod
    def get_highlighted(text):
        return '\n({0})'.format(text)

    @staticmethod
    def get_header(text, color='yellow'):
        return '\n[{0}]'.format(stylize(text.upper(), color=color))

    @staticmethod
    def get_count_styled(count, normal=0.0):
        if float(count) > normal:
            return stylize(count, color='yellow')

        return stylize(count, color='white')

    @staticmethod
    def _format_output(data):
        return '\n'.join(data)

    @staticmethod
    def get_text_table(table_header, table_body):
        return render_table(
            table_header,
            table_body
        )



# -*- coding: utf-8 -*-

from nawano.models import Alias
from ._base import NawanoService


class AliasService(NawanoService):
    __model__ = Alias

    @staticmethod
    def _get_table_body(aliases):
        for m in aliases:
            yield [
                m.name,
                '…{0}'.format(m.address[-8:]),
                m.description
            ]

    @property
    def _table_header(self):
        return ['name', 'address', 'description']

    def get_details(self, **kwargs):
        alias = self.get_one(raise_on_empty=True, **kwargs)

        return self._format_output([
            self.get_header('alias'),
            'name: ' + alias.name,
            'address: ' + alias.address,
            'description: {0}'.format(alias.description)
        ]) + '\n\n'

    def get_table(self, **kwargs):
        aliases = self.get_many(raise_on_empty=True, **kwargs)
        return self.get_text_table(self._table_header, self._get_table_body(aliases))



# -*- coding: utf-8 -*-

import uuid

from libn import account_get, account_key

from nawano.utils import to_raw, from_raw
from ._base import NawanoService


class BlockService(NawanoService):
    @property
    def receivables(self):
        for address, pending in self.__state__.pending_blocks:
            if not pending:
                continue

            for _hash, block in pending.items():
                account = self.__state__.network.get_account(address)
                previous = account['frontier'] if account else '0' * 64
                work_hash = previous if account else account_key(address)

                new_balance = (
                    str(
                        int(account['balance']) + int(block['amount'])
                    ) if account
                    else block['amount']
                )

                yield {
                    'representative': self.__state__.wallet.representative_address,
                    'balance': new_balance,
                    'link': _hash,
                    'account': address,
                    'previous': previous,
                }, work_hash

    def get_sendblock(self, account, alias_to, amount):
        address = account_get(account.public_key)
        n_account = self.__state__.network.get_account(address)
        new_balance = str(int(n_account['balance']) - int(to_raw(amount)))
        uid = uuid.uuid4()

        return {
              'representative': self.__state__.wallet.representative_address,
              'balance': new_balance,
              'id': uid.hex[:16],
              'link': account_key(alias_to.address),
              'account': address,
              'previous': n_account['frontier']
        }, n_account['frontier']

    def transaction_summary(self, account_from, alias_to, send_amount, new_balance):
        return self._format_output([
            self.get_header('new transaction', color='yellow'),
            'from: {0}/{1}'.format(account_from.name, account_get(account_from.public_key)),
            'to: {0}/{1}'.format(alias_to.name, alias_to.address),
            'amount: {0} (new balance: {1})'.format(
                send_amount,
                from_raw(new_balance)
            ) + '\n'
        ])

    def broadcast(self, block):
        return self.__state__.network.broadcast_block(block)



# -*- coding: utf-8 -*-

from nawano.models import Wallet, Account, Representative
from nawano.utils import encrypt, generate_seed, stylize

from ._base import NawanoService


class WalletService(NawanoService):
    __model__ = Wallet

    def insert(self, **kwargs):
        wallet_name = kwargs.pop('name')
        seed = kwargs.pop('seed', generate_seed())

        return self.__model__.insert(name=wallet_name, seed=encrypt(seed, kwargs.pop('password')))

    @property
    def _table_header(self):
        return ['name', 'accounts', 'available', 'pending']

    def _get_table_body(self, wallets):
        for wallet in wallets:
            accounts = Account.query(wallet_id=wallet.id).all()
            funds = self.__state__.get_wallet_funds(wallet.id)
            yield [
                wallet.name,
                len(accounts),
                funds['available'],
                funds['pending']
            ]

    def get_table(self, **kwargs):
        wallets = self.get_many(raise_on_empty=True, **kwargs)
        return self.get_text_table(self._table_header, self._get_table_body(wallets))

    @staticmethod
    def _format_rep_details(rep):
        return (
            '\n{0}name: {1}\n{0}address: {2}\n{0}weight: {3}\n{0}uptime: {4:0.2f}%\n'
            .format(' '*2, rep.alias, rep.address, rep.weight, float(rep.uptime))
        )

    def get_details(self, **kwargs):
        wallet = self.get_one(**kwargs, raise_on_empty=True)
        funds = self.__state__.get_wallet_funds(wallet.id)
        accounts = Account.query(wallet_id=wallet.id).all()

        if not wallet.representative_address:
            rep_text = stylize('\n  not set\n', color='red')
        elif isinstance(wallet.representative, Representative):
            rep_text = self._format_rep_details(wallet.representative)
        else:
            rep_text = '\n  address: {0}\n'.format(wallet.representative_address)

        return self._format_output([
            self.get_header('wallet'),
            'name: ' + wallet.name,
            'updated: ' + str(wallet.updated_on),
            'accounts: {0}'.format(len(accounts)),
            self.get_highlighted('funds') + self.funds_text(funds),
            self.get_highlighted('representative') + rep_text + '\n',
        ])



# -*- coding: utf-8 -*-

import click
from sys import stdout
from libn import account_get
from nawano.services import account_service, state_service
from nawano.status import with_status
from nawano.utils import password_input
from nawano.exceptions import NawanoError
from .root import root_group


@with_status(text='creating new account')
def _account_create(**kwargs):
    public_key = account_service.insert(**kwargs)
    account_service.refresh_balances()
    msg = 'address: {0}'.format(account_get(public_key))
    return public_key, msg


@with_status(text='deleting account')
def _account_delete(**kwargs):
    account = account_service.get_one(raise_on_empty=True, **kwargs)
    account_service.delete(**kwargs)
    account_service.refresh_balances()
    msg = 'deleted: {0}/{1}'.format(account.name, account_get(account.public_key))
    return None, msg


def _validate_account_name(ctx, param, value):
    if state_service.get_accounts(name=value):
        raise NawanoError('an account with that name already exists')

    return account_service.validate_name(value)


def _validate_account_idx(ctx, param, value):
    if not value:
        return

    account = state_service.get_accounts(idx=value)

    if account:
        raise NawanoError('this index already belongs to account: {0}'.format(account[0].name))

    return value


@root_group.group('account', short_help='account operations')
def account_group():
    assert state_service.wallet


@account_group.command('create', short_help='create account')
@click.option('--name', 'name', help='wallet name', callback=_validate_account_name, required=True)
@click.option('--index', 'idx', help='custom account index', callback=_validate_account_idx, required=False)
def account_create(**kwargs):
    kwargs['password'] = password_input(validate_confirm=False)
    _account_create(**kwargs)


@account_group.command('show', short_help='show account details')
# @click.option('--name', 'name', help='account name', required=False)
# @click.option('--address', 'address', help='account address', required=False)
# @click.option('--index', 'idx', help='account index', required=False)
@click.argument('name', required=True)
def account_show(**kwargs):
    # if not any([kwargs['name'], kwargs['address'], kwargs['idx']]):
    #     raise NawanoError('you must provide name, address or index')
    # kwargs['public_key'] = nano_account(kwargs.pop('address')) if kwargs['address'] else None
    stdout.write(account_service.get_details(**kwargs))


@account_group.command('delete', short_help='delete account')
@click.argument('name', required=True)
def account_delete(**kwargs):
    return _account_delete(**kwargs)


@account_group.command('list', short_help='list accounts')
def account_table(**kwargs):
    stdout.write(account_service.get_table(**kwargs))




# -*- coding: utf-8 -*-

import passwordmeter
from prompt_toolkit.validation import Validator, ValidationError


class PasswordQualityValidator(Validator):
    def validate(self, document):
        pw = document.text
        strength = passwordmeter.test(pw)[0] * 100
        min_strength = 80
        if strength < min_strength:
            raise ValidationError(message='strength {0:.3} < {1} ({2})'.format(strength, min_strength, 'too weak'))


class ConfirmValidator(Validator):
    def __init__(self, match, *args, **kwargs):
        self.match = match
        super(ConfirmValidator, self).__init__(*args, **kwargs)

    def validate(self, document):
        if document.text != self.match:
            raise ValidationError(message='the passwords must match')



# -*- coding: utf-8 -*-

from libn import account_key, account_get, deterministic_key

from nawano.models import Account
from nawano.exceptions import NawanoError
from nawano.utils import decrypt


from ._base import NawanoService


class AccountService(NawanoService):
    __model__ = Account

    def insert(self, **kwargs):
        wallet = self.__state__.wallet
        account_name = kwargs.pop('name')

        # Decrypt seed
        seed = decrypt(wallet.seed, kwargs.pop('password'))

        # Get next ID from input or DB
        account_idx = kwargs.pop('idx', None) or self.__model__.get_next_idx(wallet.id)

        # Derive account from seed
        sk, pk, _ = deterministic_key(seed.decode('ascii'), int(account_idx))

        # Look for existing account with this PK
        existing = self.__model__.query(public_key=pk).one_or_none()

        if existing:
            raise NawanoError('account key conflicts with {0} in wallet {1}'.format(existing.name, existing.wallet.name))

        account_pk = self.__model__.insert(
            idx=account_idx,
            name=account_name,
            public_key=pk,
            wallet_id=wallet.id,
            **kwargs
        )

        return account_pk

    def refresh_balances(self):
        for address, pending in self.__state__.pending_blocks:
            account = self.__state__.network.get_account(address)
            available_raw = account['balance'] if account else 0
            pending_raw = 0

            if pending:
                for block in pending.values():
                    pending_raw += int(block['amount'])

            self.__model__.update(
                account_key(address),
                available_raw=str(available_raw),
                pending_raw=str(pending_raw)
            )

        self.__state__.get_wallet_funds.cache_clear()

    def get_details(self, **kwargs):
        account = self.get_one(wallet_id=self.__state__.wallet.id, raise_on_empty=True, **kwargs)

        return self._format_output([
            self.get_header('account'),
            'index: ' + str(account.idx),
            'name: ' + account.name,
            'updated: ' + str(account.updated_on),
            'address: ' + account_get(account.public_key),
            'pubkey: ' + account.public_key.upper(),
            self.get_highlighted('funds') + self.funds_text({
                'available': account.available,
                'pending': account.pending,
            }),
            '\n',
        ])

    @property
    def _table_header(self):
        return ['name', 'index', 'address', 'available', 'pending']

    @staticmethod
    def _get_table_body(accounts):
        for a in accounts:
            yield [
                a.name,
                a.idx,
                '…{0}'.format(account_get(a.public_key)[-8:]),
                a.available,
                a.pending
            ]

    def get_table(self, **kwargs):
        accounts = self.__state__.get_accounts(**kwargs)

        if not accounts:
            raise NawanoError('no accounts found')

        return self.get_text_table(self._table_header, self._get_table_body(accounts))



# -*- coding: utf-8 -*-

from sqlalchemy import Column, String, Integer, DateTime, ForeignKey, UniqueConstraint
from sqlalchemy.sql import func
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from nawano.utils import from_raw

from nawano.db import Base, get_db_session
from .wallet import Wallet
from ._base import BaseMixin


class Account(Base, BaseMixin):
    __tablename__ = 'account'

    idx = Column(Integer, nullable=False, primary_key=True)
    name = Column(String, nullable=False)
    wallet_id = Column(String, ForeignKey('wallet.id'), primary_key=True)
    wallet = relationship(Wallet)

    public_key = Column(String, unique=True)
    available_raw = Column(String, nullable=False, default='0')
    pending_raw = Column(String, nullable=False, default='0')
    updated_on = Column(DateTime, default=func.now(), onupdate=func.now())

    UniqueConstraint(wallet_id, name, name='uc_wallet_account_name')

    @hybrid_property
    def available(self):
        return from_raw(self.available_raw)

    @hybrid_property
    def pending(self):
        return from_raw(self.pending_raw)

    @classmethod
    def get_next_idx(cls, wallet_id):
        with get_db_session() as s:
            curr_idx = s.query(func.max(Account.idx)).filter_by(wallet_id=wallet_id).one()[0]

        return 0 if curr_idx is None else curr_idx + 1

    @classmethod
    def query(cls, **kwargs):
        return cls._query(**kwargs)

    @classmethod
    def update(cls, public_key, **kwargs):
        existing = cls.query(public_key=public_key).one()
        return cls._update(existing, **kwargs)

    @classmethod
    def insert(cls, **kwargs):
        account = Account(**kwargs)
        cls._add(account)
        return kwargs['public_key']



# -*- coding: utf-8 -*-

import click
from prompt_toolkit import prompt
from sys import stdout
from libn import deterministic_key

from nawano.services import wallet_service, state_service, rep_service, account_service
from nawano.utils import password_input, decrypt, stylize
from nawano.status import with_status
from nawano.exceptions import NawanoError
from .root import root_group


@with_status(text='setting active wallet')
def _wallet_set_active(wallet_id):
    state_service.set_wallet(wallet_id)
    account_service.refresh_balances()
    return None, None


@with_status(text='decrypting seed')
def _dump_seed(seed_encrypted, password):
    seed = decrypt(seed_encrypted, password).decode('ascii')
    msg = seed
    return None, msg


@with_status(text='validating seed')
def _validate_seed(seed):
    if len(seed) != 64:
        raise NawanoError('must be exactly 64 chars in size')

    try:
        deterministic_key(seed, 0)
    except ValueError:
        raise NawanoError('must be a hexadecimal string')

    return seed.upper().encode('ascii'), None


@with_status(text='setting representative')
def _wallet_set_representative(address):
    if not state_service.network.validate_address(address):
        raise NawanoError('invalid address format')

    wallet_id = state_service.wallet.id
    wr = {
        'representative_address': address
    }

    rep = rep_service.get_one(address=address)

    wr['representative_alias'] = rep.alias if rep else None
    wallet_service.update(wallet_id, **wr)

    return None, None


@with_status(text='creating wallet')
def _wallet_create(**kwargs):
    wallet_id = wallet_service.insert(**kwargs)
    msg = 'local ID: {0}'.format(wallet_id)
    return wallet_id, msg


def _validate_wallet_name(ctx, param, value):
    if wallet_service.get_one(name=value):
        raise NawanoError('a wallet with that name already exists')

    return wallet_service.validate_name(value)


def _validate_rep_alias(ctx, param, value):
    if not rep_service.get_one(alias=value):
        raise NawanoError('invalid representative alias provided')

    return value


def _validate_rep_address(ctx, param, value):
    if not state_service.network.validate_address(value):
        raise NawanoError('invalid address format')

    return value


@root_group.group('wallet', short_help='wallet operations')
def wallet_group():
    pass


@wallet_group.command('create', short_help='create a new wallet')
@click.option('--name', 'name', help='wallet name', callback=_validate_wallet_name, required=True)
def wallet_create(**kwargs):
    kwargs['password'] = password_input(validate_confirm=True)
    stdout.write('\n')
    wallet_id, _ = _wallet_create(**kwargs)

    if prompt('switch to this wallet? [Y/n] ').lower() in ['', 'y']:
        _wallet_set_active(wallet_id)


@wallet_group.command('import', short_help='create wallet from existing seed')
@click.option('--name', 'name', help='wallet name', callback=_validate_wallet_name, required=True)
def wallet_import(**kwargs):
    seed = password_input(validate_confirm=False, pw1_text='seed: ')
    kwargs['seed'], _ = _validate_seed(seed.decode('ascii'))

    kwargs['password'] = password_input(validate_confirm=True)
    stdout.write('\n')
    wallet_id, _ = _wallet_create(**kwargs)

    if prompt('switch to this wallet? [Y/n] ').lower() in ['', 'y']:
        _wallet_set_active(wallet_id)


@wallet_group.command('use', short_help='set active wallet')
@click.argument('name', required=True)
def wallet_set_active(name):
    wallet = wallet_service.get_one(raise_on_empty=True, name=name)
    _wallet_set_active(wallet.id)


@wallet_group.command('show', short_help='show wallet details')
@click.argument('name', required=True)
def wallet_show(**kwargs):
    stdout.write(wallet_service.get_details(**kwargs))


@wallet_group.command('representative', short_help='set representative')
@click.argument('address', required=True)
def wallet_set_representative(address):
    return _wallet_set_representative(address)


@wallet_group.command('list', short_help='list all wallets')
def wallet_list(**kwargs):
    stdout.write(wallet_service.get_table(**kwargs))


@wallet_group.command('dump', short_help='dump seed in plain text')
@click.argument('name', required=True)
def wallet_dump_seed(name):
    stdout.write(
        '\n[{0}] the seed can be used to access your funds; store it securely!\n\n'
        'enter wallet password to proceed.\n'.format(stylize('info', color='yellow'))
    )
    password = password_input(validate_confirm=False)
    seed = wallet_service.get_one(raise_on_empty=True, name=name).seed
    return _dump_seed(seed, password)



# -*- coding: utf-8 -*-

from functools import lru_cache
from decimal import Decimal
from datetime import datetime
from libn import account_get

from nawano.clients import RPC
from nawano.models import ConfigAttribute, Account, State, Wallet
from nawano.exceptions import NoRecordsFound, NoActiveWallet


class StateService(object):
    @property
    def _state(self):
        return State.query().first()

    def get_announced(self, name):
        return getattr(self._state, self._as_announced(name))

    # @lru_cache()
    def get_active_wallet(self):
        wallet = State.get_wallet()
        if not wallet:
            raise NoActiveWallet

        return wallet

    @lru_cache()
    def get_wallet_funds(self, wallet_id):
        available = pending = Decimal(0)

        for account in Account.query(wallet_id=wallet_id).all():
            available += Decimal(account.available)
            pending += Decimal(account.pending)

        return {
            'available': str(available or 0),
            'pending': str(pending or 0)
        }

    @property
    def wallet(self):
        return self.get_active_wallet()

    @staticmethod
    def _as_announced(name):
        return '{0}_{1}'.format(name, 'announced')

    def set_announced(self, name):
        State.set(**{name + '_announced': datetime.now().replace(microsecond=0)})

    def set_wallet(self, wallet_id):
        wallet = Wallet.query(id=wallet_id).one_or_none()

        if not wallet:
            raise NoRecordsFound

        State.set(wallet_id=wallet.id)
        # self.get_active_wallet.cache_clear()

    @property
    def wallet_funds(self):
        return self.get_wallet_funds(self.wallet.id)

    @property
    def network(self):
        return RPC(ConfigAttribute.get_one('backend').value)

    def get_accounts(self, **kwargs):
        return Account.query(wallet_id=self.wallet.id, **kwargs).all()

    @property
    def pending_blocks(self):
        try:
            accounts = self.get_accounts()
        except NoActiveWallet:
            return []

        addresses = [account_get(account.public_key) for account in accounts]
        return self.network.get_pending(addresses)



# -*- coding: utf-8 -*-

import click

from decimal import Decimal
from sys import stdout
from libn import work_generate, account_get, account_key, deterministic_key, sign_block
from nawano.services import block_service, account_service, alias_service, state_service
from nawano.status import with_status
from nawano.utils import password_input, decrypt, from_raw
from nawano.exceptions import NawanoError
from .root import root_group


@with_status(text='generating proof-of-work')
def _pow_generate(_hash):
    return work_generate(_hash)


@with_status(text='refreshing balances')
def _refresh_balances():
    account_service.refresh_balances()
    return None, None


@with_status(text='preparing send-block')
def _sendblock_prepare(**kwargs):
    validated = _validate_send(kwargs)
    block, work_hash = block_service.get_sendblock(*validated)
    return (validated, block, work_hash), None


def _validate_send(payload):
    account_from = account_service.get_one(name=payload['account_from'], wallet_id=state_service.wallet.id)
    n_account = state_service.network.get_account(account_get(account_from.public_key))
    recipient_alias = alias_service.get_one(name=payload['recipient_alias'])
    send_amount = Decimal(payload['amount'])

    if not account_from:
        raise NawanoError('option --account_from must be a valid account name')
    elif not recipient_alias:
        raise NawanoError('option --recipient_alias must be a valid alias')
    elif send_amount <= 0:
        raise NawanoError('amount must be greater than 0')
    elif Decimal(from_raw(n_account['balance'])) < Decimal(send_amount):
        missing = Decimal(send_amount) - Decimal(account_from.available)
        raise NawanoError('insufficient funds (available: {0}, send_amount: {1} (missing: {2}))'
                          .format(account_from.available, send_amount, missing))

    return account_from, recipient_alias, send_amount


@with_status(text='signing transaction')
def _block_sign(block, seed):
    account = account_service.get_one(public_key=account_key(block['account']))
    assert account

    sk, pk, _ = deterministic_key(seed, account.idx)
    return sign_block(sk, pk, block)


@with_status(text='validating')
def _block_validate(block, schema):
    return schema.load(block).data


@with_status(text='broadcasting block')
def _block_broadcast(block):
    return block_service.broadcast(block)


@with_status(text='refreshing balances')
def _refresh_balances():
    return account_service.refresh_balances()


@root_group.group('funds', short_help='send or receive funds')
def funds_group():
    if not state_service.get_accounts():
        raise NawanoError('this operation requires an account')
    elif not state_service.wallet.representative_address:
        raise NawanoError('this operation requires a wallet representative')


@funds_group.command('refresh', short_help='refresh balances now')
def wallet_list():
    _refresh_balances()


@funds_group.command('send', short_help='send funds')
@click.option('--account_from', help='name of account to send from', required=True)
@click.option('--recipient_alias', help='name of recipient address alias', required=True)
@click.option('--amount', help='amount to send', required=True)
def funds_send(**kwargs):
    (validated, block, work_hash), _ = _sendblock_prepare(**kwargs)
    stdout.write(block_service.transaction_summary(*validated, block['balance']))

    stdout.write('\n---\nenter password to proceed or <ctrl+d> to cancel\n')
    password = password_input(validate_confirm=False)
    seed = decrypt(state_service.wallet.seed, password).decode('ascii')

    stdout.write('\nprocessing {0}\n'.format(block['link']))

    # Create PoW based on previous or PK
    block['work'] = _pow_generate(work_hash)

    # Sign block
    block['signature'] = _block_sign(block, seed)

    # Broadcast transaction
    _block_broadcast(block)

    # Refresh balances
    _refresh_balances()


@funds_group.command('pull', short_help='receive pending funds')
def funds_pull():
    p_tot = state_service.wallet_funds['pending']

    if float(p_tot) <= 0:
        raise NawanoError('no pending blocks')

    stdout.write(account_service.get_header('receive'))
    stdout.write('\npending: {0} ~ use <ctrl+d> to cancel\n'.format(p_tot))

    password = password_input(validate_confirm=False)
    seed = decrypt(state_service.wallet.seed, password).decode('ascii')

    for block, work_hash in block_service.receivables:
        stdout.write('\nprocessing {0}\n'.format(block['link']))

        # Create PoW based on frontier
        block['work'] = _pow_generate(work_hash)

        # Sign block
        block['signature'] = _block_sign(block, seed)

        # Broadcast transaction
        _block_broadcast(block)

        # Refresh balances
        _refresh_balances()



# -*- coding: utf-8 -*-

from sqlalchemy import Column, String, DateTime
from sqlalchemy.sql import func

from nawano.db import Base
from ._base import BaseMixin


class Alias(Base, BaseMixin):
    __tablename__ = 'alias'

    address = Column(String, nullable=False, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    description = Column(String, nullable=True)
    updated_on = Column(DateTime, default=func.now(), onupdate=func.now())

    @classmethod
    def query(cls, **kwargs):
        return cls._query(**kwargs)

    @classmethod
    def update(cls, public_key, **kwargs):
        existing = cls.query(public_key=public_key).one()
        return cls._update(existing, **kwargs)

    @classmethod
    def insert(cls, **kwargs):
        account = Alias(**kwargs)
        cls._add(account)
        return kwargs['address']



# -*- coding: utf-8 -*-


class NawanoError(Exception):
    pass


class ValidationError(NawanoError):
    pass


class UnexpectedBackendResponse(NawanoError):
    def __init__(self, message):
        self.message = message

    def __str__(self, *args, **kwargs):
        return 'backend error: {0}'.format(self.message)


class NoActiveWallet(NawanoError):
    def __str__(self, *args, **kwargs):
        return 'this operation requires an active wallet'


class NoRecordsFound(NawanoError):
    def __str__(self, *args, **kwargs):
        return 'no such object'



# -*- coding: utf-8 -*-

from contextlib import contextmanager

from sqlalchemy import create_engine, MetaData
from sqlalchemy.pool import SingletonThreadPool
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session

from nawano.settings import SQLALCHEMY_DATABASE_URI


engine = create_engine(
    SQLALCHEMY_DATABASE_URI,
    poolclass=SingletonThreadPool,
    connect_args={'check_same_thread': False}
)

Session = scoped_session(
    sessionmaker(
        bind=engine,
        autoflush=False,
        autocommit=False
    )
)


def filter_empty(data):
    return dict(filter(lambda x: x[1], data.items()))


@contextmanager
def get_db_session():
    try:
        yield Session
    finally:
        Session.remove()


Base = declarative_base()
metadata = MetaData()



# -*- coding: utf-8 -*-

import re
import click
from sys import stdout
from requests.exceptions import RequestException

from nawano.services import config_service
from nawano.status import with_status
from nawano.exceptions import NawanoError
from nawano.clients import RPC
from .root import root_group


@with_status(text='testing connectivity')
def _test_backend(uri):
    try:
        client = RPC(uri)
        client.get_supply()
    except RequestException as e:
        raise NawanoError(e)


@with_status(text='update config attribute')
def _config_attribute_set(name, value):
    config_service.set(name, value)
    updated = config_service.get(name)
    msg = '{0} is now {1}'.format(updated.name, updated.value)
    return None, msg


@root_group.command('config', short_help='manage nawano config')
@click.argument('attribute', required=False)
@click.argument('value', required=False)
def config_group(attribute=None, value=None):
    name = attribute
    if not name:
        stdout.write(config_service.table)
    elif name and value:
        if name == 'backend':
            # value = re.sub(r'https?://', '', value)
            _test_backend(value)

        _config_attribute_set(name, value)
    else:
        attr = config_service.get(name)
        stdout.write('{0}\n\n'.format(attr.value))



# -*- coding: utf-8 -*-

from sys import stdout
from datetime import datetime
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory

from click.exceptions import UsageError
from nawano.exceptions import NawanoError, NoActiveWallet

from nawano.services import state_service, config_service
from nawano.settings import PROMPT_COLOR_DEPTH, HISTORY_PATH, PROMPT_STYLE
from nawano.utils import stylize
from nawano.repl import completion, completer


def get_styled_prompt():
    return [
        ('class:prompt-name', 'nawano'),
        ('class:prompt-marker', u'> '),
    ]


def cooldown_reached(key):
    cooldown_secs = config_service.get('notify_cooldown').value
    return cooldown_secs < (datetime.now() - state_service.get_announced(key)).seconds


def pending_announce():
    try:
        funds = state_service.wallet_funds
    except NoActiveWallet:
        return

    if float(funds['pending']) > 0:
        stdout.write('[{0}] there are pending funds, enter {1} to claim now.\n'.format(
            stylize('info', color='yellow'),
            stylize('funds pull', color='green')
        ))

        state_service.set_announced('pending')


def weight_announce():
    try:
        representative = state_service.wallet.representative
        if not representative or not representative.weight:
            return
    except NoActiveWallet:
        return

    max_weight = config_service.get('max_weight').value

    if representative.weight > max_weight:
        stdout.write('[{0}] your representative has too much weight, type {1} to change\n'.format(
            stylize('info', color='yellow'),
            stylize('wallet representative', color='green')
        ))

        state_service.set_announced('weight')


def get_bottom_toolbar():
    try:
        wallet = state_service.wallet
    except NoActiveWallet:
        return None

    pending_class = (
        'class:bottom-toolbar-pending'
        if
        float(state_service.wallet_funds['pending']) > 0
        else
        'class:bottom-toolbar-value'
    )

    funds = state_service.wallet_funds

    return [
        ('class:bottom-toolbar-logo', ' Ϟ '),
        ('class:bottom-toolbar-key', 'wallet:'),
        ('class:bottom-toolbar-value', str(wallet.name)),
        ('class:bottom-toolbar-key', ' · available:'),
        ('class:bottom-toolbar-value', funds['available']),
        ('class:bottom-toolbar-key', ' · pending:'),
        (pending_class, funds['pending']),
    ]


def nawano_loop(ctx):
    session = PromptSession(
        completer=completer.NawanoCompleter(ctx),
        refresh_interval=1,
        color_depth=PROMPT_COLOR_DEPTH,
        history=FileHistory(HISTORY_PATH),
        style=PROMPT_STYLE,
    )

    while True:
        try:
            if cooldown_reached('pending'):
                pending_announce()

            if cooldown_reached('weight'):
                weight_announce()

            cmd_str = session.prompt(get_styled_prompt(), bottom_toolbar=get_bottom_toolbar)

            if not cmd_str:
                continue

        except KeyboardInterrupt:
            continue
        except EOFError:
            break

        try:
            c = completion.NawanoCompletion(ctx.command)
            c.args = cmd_str
            if c.args[0] in ['exit', 'quit']:
                break
        except ValueError as e:
            stdout.write(str(e) + '\n\n')
            continue

        try:
            with ctx.command.make_context(None, c._args, parent=ctx.parent) as ctx:
                ctx.command.invoke(ctx)
                ctx.exit()
        except (UsageError, NawanoError) as e:
            if isinstance(e, UsageError):
                msg = e.format_message()
            else:
                msg = str(e)

            stdout.write('{0} {1}\n\n'.format(stylize('>', color='red'), msg.lower()))
        except SystemExit:
            pass



# -*- coding: utf-8 -*-

import click
from sys import stdout
from nawano.services import alias_service
from nawano.status import with_status
from nawano.exceptions import NawanoError
from .root import root_group


@with_status(text='creating alias')
def _alias_create(**kwargs):
    address = alias_service.insert(**kwargs)
    msg = 'created alias {0}/{1}'.format(kwargs['name'], address)
    return address, msg


@with_status(text='deleting alias')
def _alias_delete(**kwargs):
    alias = alias_service.get_one(raise_on_empty=True, **kwargs)
    alias_service.delete(**kwargs)
    msg = 'deleted: {0}/{1}'.format(alias.name, alias.address)
    return None, msg


@with_status(text='validating address')
def _validate_address(address):
    if not alias_service.validate_address(address):
        raise NawanoError('invalid address')

    return address, None


def _validate_name(ctx, param, value):
    if alias_service.get_one(name=value):
        raise NawanoError('an alias with that name already exists')
    elif value.startswith(('xrb_', 'nano_',)):
        raise NawanoError('alias name looks like an address')

    return alias_service.validate_name(value)


def _validate_address_light(ctx, param, value):
    if not value.startswith(('xrb_', 'nano_',)):
        raise NawanoError('alias address should start with xrb_ or nano_')

    alias = alias_service.get_one(address=value)
    if alias:
        raise NawanoError('that address is already mapped to: {0}'.format(alias.name))

    return value


@root_group.group('alias', short_help='manage aliases')
def alias_group():
    pass


@alias_group.command('create', short_help='create alias')
@click.option('--name', 'name', help='alias name', callback=_validate_name, required=True)
@click.option('--address', 'address', help='alias address', callback=_validate_address_light, required=True)
@click.option('--description', 'description', help='alias description', required=False)
def alias_create(**kwargs):
    _validate_address(kwargs.get('address'))
    _alias_create(**kwargs)


@alias_group.command('show', short_help='show alias details')
# @click.option('--name', 'name', help='by name')
# @click.option('--address', 'address', help='by address')
@click.argument('name', required=True)
def alias_show(**kwargs):
    stdout.write(alias_service.get_details(**kwargs))


@alias_group.command('delete', short_help='delete alias')
@click.argument('name', required=True)
def alias_delete(**kwargs):
    # stdout.write(alias_service.get_one(**kwargs))
    return _alias_delete(**kwargs)


@alias_group.command('list', short_help='list aliases')
def alias_table(**kwargs):
    stdout.write(alias_service.get_table(**kwargs))




# -*- coding: utf-8 -*-

from .wallet import WalletService
from .block import BlockService
from .account import AccountService
from .config import ConfigService
from .alias import AliasService
from .rep import RepresentativeService
from .state import StateService

wallet_service = WalletService()
block_service = BlockService()
account_service = AccountService()
config_service = ConfigService()
alias_service = AliasService()
rep_service = RepresentativeService()
state_service = StateService()



# -*- coding: utf-8 -*-

from datetime import datetime
from nawano.db import get_db_session, filter_empty


class BaseMixin(object):
    big_bang = datetime.fromtimestamp(-2147483648)

    @classmethod
    def _query(cls, *entities, **kwargs):
        with get_db_session() as s:
            return s.query(entities or cls).filter_by(**filter_empty(kwargs))

    @classmethod
    def _update(cls, obj, **kwargs):
        with get_db_session() as s:
            for k, v in kwargs.items():
                setattr(obj, k, v)

            s.merge(obj)
            s.commit()

    @classmethod
    def delete(cls, **kwargs):
        with get_db_session() as s:
            obj = s.query(cls).filter_by(**filter_empty(kwargs)).one()
            s.delete(obj)
            s.commit()

    @classmethod
    def _add(cls, obj):
        with get_db_session() as s:
            s.add(obj)
            s.commit()



# -*- coding: utf-8 -*-

import click

from sys import stdout
from nawano import REPL_LOGO
from nawano.repl.loop import nawano_loop
from nawano.utils import stylize
from nawano.services import config_service, account_service, rep_service, state_service
from nawano.task import Task


def tasks_start(tasks_args):
    tasks = []
    for task_args in tasks_args:
        task = Task(*task_args)
        task.start()
        tasks.append(task)

    return tasks


def get_cfg(key):
    return config_service.get(key).value


@click.group(invoke_without_command=True)
@click.pass_context
def root_group(ctx):
    if ctx.invoked_subcommand is None:
        stdout.write('\n'.join(REPL_LOGO))

        tab_hint = stylize('TAB', color='yellow')
        stdout.write('use the :{0}: key for auto-completion\n\n'.format(tab_hint))

        state_service.is_syncing = False

        # Start worker threads
        tasks = tasks_start(
            [
                [account_service.refresh_balances, get_cfg('balance_refresh')],
                [rep_service.refresh_reps, get_cfg('reps_refresh') * 60]
            ]
        )

        # Start REPL
        nawano_loop(ctx)

        # Stop worker threads
        for task in tasks:
            task.stop()

        stdout.write('\nnawano stopping.. buh-bye!\n')


@root_group.command('exit', short_help='exit nawano')
def exit_nawano():
    pass



# -*- coding: utf-8 -*-

from sqlalchemy import Column, String, Integer, DateTime, Float
from sqlalchemy.sql import func

from nawano.db import Base
from ._base import BaseMixin


class Representative(Base, BaseMixin):
    __tablename__ = 'representative'

    alias = Column(String, primary_key=True)
    address = Column(String, primary_key=True)
    uptime = Column(String, nullable=False)
    delegators = Column(Integer, nullable=False)
    weight = Column(Float, nullable=False)
    updated_on = Column(DateTime, default=func.now(), onupdate=func.now(), nullable=False)

    @classmethod
    def query(cls, **kwargs):
        return cls._query(**kwargs)



# -*- coding: utf-8 -*-

import privy
import crayons

from os import urandom
from sys import stdout
from texttable import Texttable
from prompt_toolkit import prompt
from libn import mrai_from_raw, mrai_to_raw

from nawano.settings import PW_MESSAGE_INFO
from nawano.validators import PasswordQualityValidator
from nawano.exceptions import NawanoError


def stylize(text, color='white', **kwargs):
    return getattr(crayons, color)(text, **kwargs).__str__()


def from_raw(amount):
    mrai = mrai_from_raw(amount)
    return mrai.rstrip('0') if float(mrai) else 0


def to_raw(amount):
    return mrai_to_raw(str(amount))


def render_table(header, body):
    t = Texttable(max_width=110)
    t.set_cols_dtype(['t'] * len(header))
    t.set_header_align(['l'] * len(header))
    t.set_deco(Texttable.HEADER | Texttable.VLINES)
    t.header(header)
    for row in body:
        t.add_row(row)

    return '\n' + t.draw() + '\n\n'


def password_input(validate_confirm=False, pw1_text='password: ', pw2_text='confirm: '):
    qv = PasswordQualityValidator() if validate_confirm else None
    if qv:
        pre_prompt = 'enter the desired password to proceed or <ctrl+d> to cancel'
        stdout.write('\n{0}\n\n{1}\n'.format('\n'.join(PW_MESSAGE_INFO), pre_prompt))

    try:
        password1 = prompt(pw1_text, is_password=True, validator=qv).encode('utf-8')

        if not validate_confirm:
            return password1

        password2 = prompt(pw2_text, is_password=True).encode('utf-8')

        if password1 != password2:
            raise NawanoError('password confirmation failed')

        return password1
    except (KeyboardInterrupt, EOFError):
        raise NawanoError('aborted')


def generate_seed():
    seed = urandom(32).hex().upper()
    return seed.encode('ascii')


def encrypt(data, password):
    return privy.hide(data, password)


def decrypt(encrypted, password):
    try:
        return privy.peek(encrypted, password)
    except ValueError:
        raise NawanoError('invalid wallet password')



# -*- coding: utf-8 -*-

from uuid import uuid4
from sqlalchemy import Column, String, DateTime, ForeignKeyConstraint
from sqlalchemy.sql import func
from sqlalchemy.orm import relationship

from nawano.db import Base

from .rep import Representative
from ._base import BaseMixin


class Wallet(Base, BaseMixin):
    __tablename__ = 'wallet'

    id = Column(String, primary_key=True, nullable=False)
    name = Column(String, unique=True, nullable=False)
    seed = Column(String, nullable=False)
    updated_on = Column(DateTime, default=func.now(), onupdate=func.now())

    representative_address = Column(String)
    representative_alias = Column(String)
    representative = relationship(Representative)

    ForeignKeyConstraint(
        (representative_address, representative_alias),
        (Representative.address, Representative.alias)
    ),

    @classmethod
    def query(cls, **kwargs):
        return cls._query(**kwargs)

    @classmethod
    def update(cls, wallet_id, **kwargs):
        wallet = cls.query(id=wallet_id).one()
        res = cls._update(wallet, **kwargs)
        return res

    @classmethod
    def insert(cls, **kwargs):
        kwargs['id'] = str(uuid4())
        cls._add(Wallet(**kwargs))
        return kwargs['id']



# -*- coding: utf-8 -*-

from sqlalchemy import Column, String, Integer, DateTime, ForeignKey
from sqlalchemy.orm import relationship

from nawano.db import Base
from nawano import __version__
from ._base import BaseMixin
from .wallet import Wallet


class State(Base, BaseMixin):
    __tablename__ = 'state'

    id = Column(Integer, primary_key=True)

    version = Column(String, default=__version__)

    wallet_id = Column(String, ForeignKey('wallet.id'))
    wallet = relationship(Wallet, foreign_keys=wallet_id)

    pending_announced = Column(DateTime, default=BaseMixin.big_bang, nullable=False)
    weight_announced = Column(DateTime, default=BaseMixin.big_bang, nullable=False)

    @classmethod
    def query(cls, **kwargs):
        return cls._query(**kwargs)

    @classmethod
    def install(cls):
        return cls._add(State(id=1))

    @classmethod
    def get_wallet(cls):
        return cls.query().first().wallet

    @classmethod
    def set(cls, **kwargs):
        return cls._update(cls.query().first(), **kwargs)



# -*- coding: utf-8 -*-

import re
import shlex
import click
from prompt_toolkit.completion import Completion

from nawano.services import wallet_service, config_service, alias_service, rep_service, state_service
from nawano.exceptions import NoActiveWallet
from nawano.settings import RPC_BACKENDS


class NawanoCompletion(object):
    _args_text = None
    _args = None

    def __init__(self, ctx):
        self._ctx = ctx

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, args_text):
        self._args_text = args_text
        self._args = shlex.split(args_text)

    @property
    def _is_within_command(self):
        return self._args_text.rstrip() == self._args_text

    @property
    def req_backends(self):
        return self._args[:2] == ['config', 'backend']

    @property
    def req_reps(self):
        if len(self.args) != 2:
            return False

        return self._args[:2] == ['wallet', 'representative']

    @property
    def req_wallets(self):
        if len(self._args) != 2:
            return False

        return self._args[0] == 'wallet' and self._args[1] in ['use', 'show', 'dump']

    @property
    def req_accounts(self):
        if len(self._args) != 2:
            return False

        return self._args[0] == 'account' and self._args[1] in ['show', 'delete']

    @property
    def req_aliases(self):
        if len(self._args) < 2:
            return False

        if self._args[0] == 'alias' and self._args[1] in ['show', 'delete']:
            return True
        elif self._args[:2] == ['funds', 'send'] and self._args[-1] == '--recipient_alias':
            return True

        return False

    @property
    def req_accounts_balances(self):
        return self._args[:2] == ['funds', 'send'] and self._args[-1] == '--account_from'

    @property
    def req_config(self):
        if len(self.args) > 1:
            return False

        return self._args[:1] == ['config']

    def _reps_formatted(self, reps):
        suggestions = self._suggest_from_objs(
            reps,
            text_key='address',
            meta_str='{weight}',
            display_key='alias'
        )

        for value, display, meta in suggestions:
            yield value, display, 'weight: {0:.3}'.format(meta)

    @property
    def _suggestions(self):
        if self.req_accounts or self.req_accounts_balances:
            try:
                accounts = state_service.get_accounts()
            except NoActiveWallet:
                return [['', 'requires an active wallet', None]]
            else:
                if accounts:
                    return self._suggest_from_objs(accounts, meta_str='{available}')
                else:
                    return [['', 'no accounts found', None]]
        elif self.req_backends:
            return RPC_BACKENDS
        elif self.req_aliases:
            aliases = alias_service.get_many()
            return self._suggest_from_objs(aliases, meta_str='{description}')
        elif self.req_wallets:
            wallets = wallet_service.get_many()
            return self._suggest_from_objs(wallets)
        elif self.req_config:
            config = config_service.get_many()
            return self._suggest_from_objs(config)
        elif self.req_reps:
            reps = rep_service.get_many()
            return self._reps_formatted(reps)

        return False

    def _get_options(self, param):
        for opt in param.opts:
            if opt in self.args:
                continue

            yield opt, None, param.help

    @staticmethod
    def _get_arguments(param):
        for opt in param.type.choices:
            yield opt, None, None

    def _get_commands(self, command):
        for c in command.list_commands(command):
            cmd = command.get_command(self._ctx, c)
            yield cmd.name, None, cmd.short_help

    @staticmethod
    def _suggest_from_objs(objs, text_key='name', meta_str=None, display_key=None):
        def matches():
            return [match.group(1) for match in re.finditer(r'{(.*?)}', meta_str or '', re.DOTALL)]

        for obj in objs:
            meta_keys = {m: getattr(obj, m) for m in matches()}
            meta = meta_str.format(**meta_keys) if meta_keys else None
            display = getattr(obj, display_key) if display_key else None
            value = getattr(obj, text_key)
            yield (value, display, meta)

    def _suggest_from_click(self, command):
        if not command:
            return

        for param in command.params:
            if param.opts[0] == self.args[-1]:
                return

        for param in command.params:
            if isinstance(param, click.Option):
                yield from self._get_options(param)
            elif isinstance(param, click.Argument) and isinstance(param.type, click.Choice):
                yield from self._get_arguments(param)
        if isinstance(command, click.MultiCommand):
            yield from self._get_commands(command)

    def get_matches(self, query, command):
        suggestions = self._suggestions or self._suggest_from_click(command)
        for text, display, meta in suggestions:
            if not text.lower().startswith(query.lower()):
                continue

            start = -len(query) if text else -1

            yield Completion(text, start, display_meta=meta, display=display)



# -*- coding: utf-8 -*-

from sys import argv, exit, stdout, version_info
from os import path, makedirs
from nawano.utils import stylize
from nawano.models import meta_create_all, State
from nawano.commands import cli
from nawano.status import with_status
from nawano.settings import NAWANO_HOME, HISTORY_PATH, DEFAULT_CORE_SETTINGS
from nawano.services import config_service


@with_status(text='creating home')
def _home_create():
    makedirs(NAWANO_HOME)


@with_status(text='touching history')
def _history_touch():
    open(HISTORY_PATH, 'w').close()


@with_status(text='seeding database')
def _database_seed():
    meta_create_all()
    State.install()
    for attr in DEFAULT_CORE_SETTINGS:
        config_service.insert(**attr)


def run():
    if version_info < (3, 6):
        exit('error: nawano requires Python 3.6 or greater')

    if len(argv) > 1:
        exit("error: nawano doesn't take arguments")

    if not path.exists(NAWANO_HOME):
        stdout.write('\n## first-time run, installing in {0}\n'
                     .format(stylize(NAWANO_HOME, color='white', bold=True)))

        _home_create()
        _history_touch()
        _database_seed()

    # Invoke CLI
    cli(prog_name='nawano')


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

from functools import wraps
from sys import stdout
from nawano.utils import stylize
from nawano.exceptions import NawanoError


class StatusMessage(object):
    l_just = 7

    def __init__(self, text):
        self._text = text

    def _create_message(self, status, color, end='\n', result=None):
        msg = '\r[{0}] {1}{2}'.format(
            stylize(status, color=color).ljust(self.l_just),
            self._text,
            end
        )

        if result is None:
            return msg

        return '{0}{1} {2}\n\n'.format(msg, stylize('>', color=color), result)

    def get_completed(self, **kwargs):
        return self._create_message('success', 'green', **kwargs)

    def get_failed(self, **kwargs):
        return self._create_message('failure', 'red', **kwargs)

    def get_working(self):
        return self._create_message('working', 'yellow', end='')


def with_status(text):
    status = StatusMessage(text)

    def inner(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                stdout.write(status.get_working())
                res = f(*args, **kwargs)
                msg = res[1] if isinstance(res, tuple) else None
                stdout.write(status.get_completed(result=msg))

                return res
            except Exception as exc:
                if isinstance(exc, NawanoError):
                    stdout.write(status.get_failed(result=str(exc)))
                    raise SystemExit

                raise

        return wrapped

    return inner



# -*- coding: utf-8 -*-

import click

from prompt_toolkit.completion import Completer
from .completion import NawanoCompletion


class NawanoCompleter(Completer):
    def __init__(self, ctx):
        self._ctx = ctx
        self._text_before = None
        self._completion = NawanoCompletion(ctx.command)

    @property
    def _cmd_ctx(self):
        ctx = self._ctx.command.make_context('', self._completion.args, resilient_parsing=True)

        while ctx.protected_args + ctx.args and isinstance(ctx.command, click.MultiCommand):
            a = ctx.protected_args + ctx.args
            cmd = ctx.command.get_command(ctx, a[0])

            if cmd is None:
                return None

            ctx = cmd.make_context(a[0], a[1:], parent=ctx, resilient_parsing=True)

        return ctx

    @property
    def _command(self):
        return self._cmd_ctx.command if self._cmd_ctx else None

    @property
    def _query(self):
        return self._completion.args.pop() if self._completion.args and self._completion._is_within_command else ''

    def get_completions(self, document, complete_event=None):
        self._text_before = document.text_before_cursor

        try:
            self._completion.args = self._text_before
        except ValueError:
            # Invalid input, often caused by missing closing quotation.
            return []

        return self._completion.get_matches(self._query, self._command)



# -*- coding: utf-8 -*-

import json
import requests

from simplejson.scanner import JSONDecodeError
from nawano.exceptions import UnexpectedBackendResponse


class RPC(object):
    def __init__(self, uri):
        self._uri = uri
        self._session = requests.session()

    def _request(self, payload):
        try:
            response = self._session.post(url=self._uri, json=payload).json()
        except JSONDecodeError:
            raise UnexpectedBackendResponse(message='unexpected response from server')

        if 'error' in response:
            raise UnexpectedBackendResponse(message=response['error'])

        return response

    def _get_blocks(self, payload):
        response = self._request(payload)
        return response['blocks'] if 'blocks' in response else {}

    def validate_address(self, address):
        response = self._request({'action': 'validate_account_number', 'account': address})
        return response['valid'] == '1'

    def broadcast_block(self, block):
        block['type'] = 'state'
        return self._request({'action': 'process', 'block': json.dumps(block)})

    def get_pending(self, addresses):
        try:
            blocks = self._get_blocks({'action': 'accounts_pending', 'source': 'true', 'accounts': addresses})
            return blocks.items() if blocks else {}
        except UnexpectedBackendResponse as err:
            if not err.message == 'Account not found':
                raise

            return {}

    def get_supply(self):
        return self._request({'action': 'available_supply'})

    def get_representatives(self):
        return self._request({'action': 'representatives'})

    def get_account(self, address):
        payload = {
            'action': 'account_info',
            'account': address,
            'representative': 'true',
            'weight': 'true',
        }

        try:
            return self._request(payload)
        except UnexpectedBackendResponse as err:
            if not err.message == 'Account not found':
                raise

            return None

    def get_blocks_info(self, hashes):
        return self._get_blocks({'action': 'blocks_info', 'hashes': hashes})

    def get_balances(self, addresses):
        try:
            return self._request({'action': 'accounts_balances', 'accounts': addresses}).get('balances')
        except UnexpectedBackendResponse as err:
            if not err.message == 'Account not found':
                raise

            return None



# -*- coding: utf-8 -*-

from functools import lru_cache

from nawano.models import ConfigAttribute
from nawano.exceptions import NawanoError

from ._base import NawanoService


class ConfigService(NawanoService):
    __model__ = ConfigAttribute

    @property
    def _table_header(self):
        return ['name', 'value', 'description']

    @property
    def table(self):
        return self.get_text_table(self._table_header, self._table_body)

    @property
    def _table_body(self):
        def truncate(value):
            if len(value) > 25:
                return '{0}…'.format(value[:25])

            return value

        for c in self.get_many():
            yield [
                c.name,
                truncate(c.value),
                c.description
            ]

    @lru_cache()
    def get(self, key):
        attr = self.__model__.get_one(key)
        if not attr:
            raise NawanoError('unrecognized attribute')

        return attr

    def set(self, key, value):
        attr = self.get(key)

        try:
            eval(attr.type)(value)
        except ValueError:
            raise NawanoError('unrecognized type provided for {0}'.format(attr.name))

        self.__model__.update(name=key, value=value)
        self.get.cache_clear()



# -*- coding: utf-8 -*-

from nawano.db import Base, engine

from .account import Account
from .wallet import Wallet
from .config import ConfigAttribute
from .alias import Alias
from .rep import Representative
from .state import State


def meta_create_all():
    Base.metadata.create_all(engine)



# -*- coding: utf-8 -*-

import sys
import os
import io
import ast

from setuptools import setup, find_packages


def get_version():
    with io.open('nawano/__init__.py') as input_file:
        for line in input_file:
            if line.startswith('__version__'):
                return ast.parse(line).body[0].value.s


if sys.platform == 'linux':
    # override nanopy/libn defaults for OSX
    os.environ['USE_GPU'] = os.environ.get('USE_GPU', '1')
    os.environ['LINK_OMP'] = os.environ.get('LINK_OMP', '1')


with io.open('README.md') as readme:
    setup(
        name='nawano',
        packages=find_packages(),
        version=get_version(),
        description='Nano REPL CLI',
        entry_points={
            'console_scripts': [
                'nawano=nawano.cli:run',
            ]
        },
        long_description=readme.read(),
        install_requires=[
            'privy==6.0.0',
            'pyblake2==1.1.2',
            'passwordmeter==0.1.8',
            'simplejson>=3.11.1',
            'click>=6.7',
            'prompt-toolkit>=2.0.4',
            'sqlalchemy>=1.2.7',
            'sqlalchemy-utils>=0.33.3',
            'requests>=2.18.4',
            'texttable>=1.4.0',
            'libn>=0.1.7',
            'crayons>=0.1.2'  # @TODO - remove
        ],
        author='Robert Wikman',
        author_email='rbw@vault13.org',
        maintainer='Robert Wikman',
        maintainer_email='rbw@vault13.org',
        python_requires='>=3.6',
        url='https://github.com/rbw/nawano',
        download_url='https://github.com/rbw/nawano/tarball/%s' % get_version(),
        keywords=['nano', 'cryptocurrency', 'cli', 'wallet'],
        platforms='any',
        classifiers=[
            'Topic :: Utilities',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Software Development :: Libraries :: Python Modules'
        ],
        license='BSD-2-Clause',
    )



# -*- coding: utf-8 -*-

import asyncio

from contextlib import suppress
from threading import Thread


class Task(Thread):
    def __init__(self, func, interval, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._func = func
        self._interval = interval
        self._loop = None
        self._poll_task = None

    def run(self):
        self._loop = asyncio.new_event_loop()
        loop = self._loop
        asyncio.set_event_loop(loop)

        try:
            # create
            self._poll_task = asyncio.ensure_future(self._poll())

            # run
            loop.run_forever()
            loop.run_until_complete(loop.shutdown_asyncgens())

            # cancel
            self._poll_task.cancel()
            with suppress(asyncio.CancelledError):
                loop.run_until_complete(self._poll_task)
        finally:
            loop.close()

    def stop(self):
        self._loop.call_soon_threadsafe(self._loop.stop)

    async def _poll(self):
        while True:
            self._loop.call_soon_threadsafe(self._func)
            await asyncio.sleep(self._interval)



# -*- coding: utf-8 -*-

from sqlalchemy import String, Column

from nawano.db import Base
from ._base import BaseMixin


class ConfigAttribute(Base, BaseMixin):
    __tablename__ = 'config'

    name = Column(String, primary_key=True, nullable=False)
    value = Column(String, nullable=False)
    description = Column(String)
    type = Column(String)

    @classmethod
    def query(cls, **kwargs):
        return cls._query(**kwargs)

    @classmethod
    def update(cls, name, **kwargs):
        existing = cls.query(name=name).one()
        return cls._update(existing, value=kwargs['value'])

    @classmethod
    def insert(cls, **kwargs):
        attr = ConfigAttribute(**kwargs)
        return cls._add(attr)

    @classmethod
    def get_one(cls, name):
        attr = cls.query(name=name).one_or_none()
        if not attr:
            return None

        attr.value = eval(attr.type)(attr.value)
        return attr



# -*- coding: utf-8 -*-

from os import path
from prompt_toolkit.output.color_depth import ColorDepth
from prompt_toolkit.styles import Style

NAWANO_HOME = '{0}/.nawano'.format(path.expanduser('~'))
DB_PATH = '{0}/database'.format(NAWANO_HOME)
HISTORY_PATH = '{0}/history'.format(NAWANO_HOME)

# Nawano database URI
SQLALCHEMY_DATABASE_URI = 'sqlite+pysqlite:///{0}'.format(DB_PATH)

# URL for obtaining list of representatives
REPRESENTATIVES_URI = 'https://mynano.ninja/api/accounts/active'

PROMPT_COLOR_DEPTH = ColorDepth.ANSI_COLORS_ONLY
PROMPT_STYLE = Style([
    ('completion-menu', 'bg:ansiyellow fg:ansiblack'),
    ('completion-menu.completion.current', 'fg:ansiwhite'),
    ('completion-menu.meta.completion', 'bg:ansibrightblack fg:ansiwhite'),
    ('completion-menu.meta.completion.current', 'fg:ansiwhite'),
    ('prompt-name', 'fg:ansiwhite'),
    ('prompt-marker', 'fg:ansigreen'),
    ('validation_error', 'bg: ansiblue'),
    ('bottom-toolbar-key', 'bg:ansigray'),
    ('bottom-toolbar-value', 'bg:ansiwhite'),
    ('bottom-toolbar-pending', 'bg:ansiyellow'),
    ('bottom-toolbar-logo', 'bg:ansigreen'),
    ('bottom-toolbar', 'fg:ansiblack bg:ansiwhite'),
])

RPC_BACKENDS = [
    ('https://getcanoe.io/rpc', 'canoe', 'canoe RPC backend'),
    ('http://[::1]:7076', 'local IPv6', 'local RPC on port 7076'),
    ('http://127.0.0.1:7076', 'local IPv4', 'local RPC on port 7076'),
]

PW_MESSAGE_INFO = [
    '- Use a good mix of numbers, letters, and symbols',
    '- Avoid using one of the ten thousand most common passwords',
    '- Use a good mix of UPPER case and lower case letters)'
]

DEFAULT_CORE_SETTINGS = [
    {
        'name': 'notify_cooldown',
        'value': '30',
        'description': 'notification cooldown',
        'type': 'int'
    },
    {
        'name': 'balance_refresh',
        'value': '10',
        'description': 'balance refresh interval',
        'type': 'int'
    },
    {
        'name': 'reps_refresh',
        'value': '50',
        'description': 'representatives refresh interval',
        'type': 'int'
    },
    {
        'name': 'max_weight',
        'value': '4.0',
        'description': 'representative max weight',
        'type': 'float'
    },
    {
        'name': 'backend',
        'value': 'https://getcanoe.io/rpc',
        'description': 'node network relay',
        'type': 'str'
    },
]


ADDRESS_PREFIX = 'xrb'



# -*- coding: utf-8 -*-

import requests
from simplejson.scanner import JSONDecodeError

from nawano.models import Representative, ConfigAttribute
from nawano.settings import REPRESENTATIVES_URI
from nawano.db import get_db_session

from ._base import NawanoService


class RepresentativeService(NawanoService):
    __model__ = Representative

    def get_many(self, **kwargs):
        max_weight = ConfigAttribute.get_one(name='max_weight').value

        return (
            self.__model__.query(**kwargs)
                .filter(self.__model__.weight < max_weight * 0.9)
                .order_by(Representative.weight.asc())
                .order_by(Representative.uptime.desc())
                .all()
        )

    def refresh_reps(self):
        # @TODO - look into bulk upserts with SQA
        try:
            representatives = requests.get(REPRESENTATIVES_URI).json()
        except JSONDecodeError:
            # @TODO - log error
            return False

        with get_db_session() as s:
            for rep in representatives:
                if 'alias' not in rep:
                    continue

                rep['weight'] = str(rep.pop('votingweight'))[:4]
                rep['address'] = rep.pop('account')

                existing = self.get_one(address=rep['address'])

                if not existing:
                    rep = Representative(**rep)
                    s.add(rep)
                else:
                    for k, v in rep.items():
                        setattr(existing, k, v)

                    s.merge(existing)

                s.flush()

            s.commit()

        # self.__state__.get_active_wallet.cache_clear()



# -*- coding: utf-8 -*-

from .root import root_group
from .account import account_group
from .wallet import wallet_group
from .config import config_group
from .alias import alias_group
from .funds import funds_group

cli = root_group
